from idp_authentication import SessionPort
from idp_authentication.users.domain.ports.unit_of_work import UnitOfWorkPort


class BaseUnitOfWork(UnitOfWorkPort):
    def __init__(self, session: SessionPort):
        self.session = session

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # If anything is left in the session, rollback
        self.rollback()
        self.session.close()

    def commit(self):
        self.session.commit()

    def rollback(self):
        self.session.rollback()



from idp_authentication.exceptions import UserNotFound
from idp_authentication.users.base_classes.base_use_case import UseCasePort
from idp_authentication.users.domain.entities.user import User
from idp_authentication.users.domain.ports.unit_of_work import UsersUnitOfWorkPort


class GetUserUseCase(UseCasePort):
    """
    Get user by idp_user_id.
    """

    def __init__(self, users_unit_of_work: UsersUnitOfWorkPort):
        self.users_unit_of_work = users_unit_of_work

    def execute(self, /, username: str) -> User:
        """Get user by username."""
        user: User = self.users_unit_of_work.user_repository.get_first(
            username=username
        )
        if user is None:
            raise UserNotFound(username)

        return user



from idp_authentication.users.base_classes.base_repository import BaseRepository
from idp_authentication.users.domain.entities import UserRole
from idp_authentication.users.domain.ports.repository import UserRoleRepositoryPort


class UserRoleRepository(BaseRepository, UserRoleRepositoryPort):
    entity = UserRole



from sqlalchemy.event import listen
from sqlalchemy.exc import ArgumentError
from sqlalchemy.orm import mapper as orm_mapper
from sqlalchemy.orm import relationship

from idp_authentication.users.base_classes.base_use_case import UseCasePort
from idp_authentication.users.domain.entities import AppEntity, User, UserRole
from idp_authentication.users.infrastructure.database.sqlalchemy.user import UserModel
from idp_authentication.users.infrastructure.database.sqlalchemy.user_role import (
    UserRoleModel,
)


class AppEntitySignal:
    """
    AppEntitySignal is a class that listens to the events of the AppEntity class
    and calls the use case to execute the business logic.

    The use case is injected into the class via the constructor.

    Parameters
    ----------
    use_case : UseCasePort
        The use case to be called when the event is triggered.

    Methods
    -------
    receive_after_insert_or_update(mapper, connection, target)
        The method that is called when the AppEntity class is inserted or updated.

    receive_after_delete(mapper, connection, target)
        The method that is called when the AppEntity class is deleted.

    """

    def __init__(self, use_case):
        self.use_case: UseCasePort = use_case

    def receive_after_insert_or_update(
        self,
        mapper,
        connection,
        target: AppEntity,
    ):
        print("received insert/update event for", target.entity_type)
        self.use_case.execute(target)

    def receive_after_delete(
        self,
        mapper,
        connection,
        target: AppEntity,
    ):
        print("received delete event for", target.entity_type)
        self.use_case.execute(app_entity_record=target, deleted=True)


def start_mappers(send_event_use_case):
    """
    Starts mappers for User and UserRole entities between SQLAlchemy models and domain entities.

    Parameters
    ----------
    send_event_use_case : UseCasePort
        Use case to send events(signals)

    Returns
    -------
    None
    """
    try:
        print("Starting Users module mappers...")

        user_mapper = orm_mapper(
            User,
            UserModel.__table__,
            properties={
                "user_roles": relationship(
                    UserRole, back_populates="user", lazy="select"
                ),
            },
        )
        orm_mapper(
            UserRole,
            UserRoleModel.__table__,
            properties={
                "user": relationship(
                    user_mapper, back_populates="user_roles", uselist=False
                ),
            },
        )
        app_entity_signal = AppEntitySignal(send_event_use_case)
        listen(
            AppEntity,
            "after_insert",
            app_entity_signal.receive_after_insert_or_update,
            propagate=True,
        )
        listen(
            AppEntity,
            "after_update",
            app_entity_signal.receive_after_insert_or_update,
            propagate=True,
        )
        listen(
            AppEntity,
            "after_delete",
            app_entity_signal.receive_after_delete,
            propagate=True,
        )
    except ArgumentError as e:
        # Skip if mappers already started
        if "already has a primary mapper defined" not in str(e):
            raise e



from typing import Any

from idp_authentication.users.base_classes.base_repository import BaseRepository
from idp_authentication.users.domain.entities.user import User
from idp_authentication.users.domain.entities.user_role import UserRole
from idp_authentication.users.domain.ports.repository import UserRepositoryPort


class UserRepository(UserRepositoryPort, BaseRepository):
    entity = User

    def get_users_with_access_to_records(
        self,
        app_entity_type: str,
        record_identifier: Any,
        roles: list[str],
    ):
        """Get users with access to records."""

        user_roles_data = (
            self.session.query(UserRole.user_id, UserRole.app_entities_restrictions)
            .filter(UserRole.role.in_(roles))
            .all()
        )

        user_ids = []
        for user_id, app_entities_restrictions in user_roles_data:
            if app_entities_restrictions is None:
                user_ids.append(user_id)
            elif app_entity_type in app_entities_restrictions:
                app_entity_restriction = app_entities_restrictions[app_entity_type]
                if (
                    app_entity_restriction is None
                    or record_identifier in app_entity_restriction
                ):
                    user_ids.append(user_id)

        return (
            self.session.query(User)
            .filter(
                User.is_active.is_(True),
                User.id.in_(user_ids),
            )
            .all()
        )

    def remove_user_role(self, user: User, role: UserRole):
        user.user_roles.remove(role)



from idp_authentication.enums import ChoiceEnum
from idp_authentication.users.base_classes.base_use_case import BaseUseCase


class GetAllowedRecordIdentifiersUseCase(BaseUseCase):
    """Get allowed app entity records identifiers use case."""

    def execute(
        self, user, role: ChoiceEnum, app_entity_type: str, permission: str = None
    ):
        return self._get_allowed_app_entity_records_identifiers(
            user=user,
            role=role,
            app_entity_type=app_entity_type,
            permission=permission,
        )



import uuid
from abc import ABC
from dataclasses import dataclass, field
from datetime import datetime


@dataclass
class BaseEntity(ABC):
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)



from abc import ABC, abstractmethod

from idp_authentication.users.domain.ports.repository import (
    UserRepositoryPort,
    UserRoleRepositoryPort,
)
from idp_authentication.users.domain.ports.session import SessionPort


class UnitOfWorkPort(ABC):
    session: SessionPort

    def __enter__(self):
        """To implement"""

    def __exit__(self, *args):
        """To implement"""

    @abstractmethod
    def commit(self):
        """Commit"""

    @abstractmethod
    def rollback(self):
        """Rollback"""


class UsersUnitOfWorkPort(UnitOfWorkPort, ABC):
    user_repository: UserRepositoryPort
    user_role_repository: UserRoleRepositoryPort



from typing import Any, List

from idp_authentication.custom_types import ALL
from idp_authentication.enums import ChoiceEnum
from idp_authentication.exceptions import AccessDenied
from idp_authentication.users.base_classes.base_use_case import BaseUseCase


class AuthorizeEntityRecordsUseCase(BaseUseCase):
    def execute(
        self,
        user,
        role: ChoiceEnum,
        app_entity_type: str,
        app_entity_records_identifiers: List[Any],
        permission: str = None,
    ):
        """
        Verify if the user has access to the requested entity records.
        If a permission is specified, and it has restrictable=True, the access is verified based on it.
        Args:
            user:                           The user performing the request
            role:                           The role that the user is acting as.
            app_entity_type:                The app entity being accessed
            app_entity_records_identifiers: The identifiers of the records belonging to the specified app entity
            permission:                     In case of specific permissions we can have permission restrictions
                                                through IDP. The value is the name of the permission
        Raises:
            AccessDenied: In case the requested records are not allowed
        """

        allowed_app_entity_records_identifiers = (
            self._get_allowed_app_entity_records_identifiers(
                user=user,
                role=role,
                app_entity_type=app_entity_type,
                permission=permission,
            )
        )

        if allowed_app_entity_records_identifiers == ALL:
            return

        if not set(app_entity_records_identifiers).issubset(
            set(allowed_app_entity_records_identifiers)
        ):
            raise AccessDenied(
                "You are not allowed to access the records in the requested entity!"
            )



from idp_authentication.users.domain.ports.event_producer import EventProducerPort


class MockEventProducer(EventProducerPort):
    def __init__(self):
        self.topic = None
        self.key = None
        self.value = None

    def produce(self, topic: str, key: str, value: dict):
        self.topic = topic
        self.key = key
        self.value = value



import abc
from typing import Any, Literal

from idp_authentication.custom_types import ALL
from idp_authentication.enums import ChoiceEnum
from idp_authentication.users.domain.ports.unit_of_work import UsersUnitOfWorkPort


class UseCasePort(abc.ABC):
    @abc.abstractmethod
    def execute(self, *args, **kwargs):
        """Execute the use case."""


class BaseUseCase(UseCasePort):
    def __init__(
        self,
        users_unit_of_work: UsersUnitOfWorkPort,
        roles: ChoiceEnum,
        app_entity_types: list[str],
    ):
        self.users_unit_of_work = users_unit_of_work
        self.roles = roles
        self.app_entity_types = app_entity_types

    @abc.abstractmethod
    def execute(self, *args, **kwargs):
        """Execute the use case."""

    def _verify_app_entity_type(self, app_entity_type: str):
        assert (
            app_entity_type in self.app_entity_types
        ), f"App Entity type does not exist: {app_entity_type}"

    def _verify_role(self, role: ChoiceEnum):
        assert role in self.roles, f"Role does not exist: {role}"

    def _get_allowed_app_entity_records_identifiers(
        self,
        user,
        role: ChoiceEnum,
        app_entity_type: str,
        permission: str = None,
    ) -> list[Any] | Literal[ALL]:
        """
        Gets the identifiers of the app entity records that the user can access.
        If no restriction both on role and permission level, return '__all__'
        Args:
            user:               The user performing the request
            role:               The role that the user is acting as.
            app_entity_type:    The app entity being accessed
            permission:         In case of specific permissions we can have permission restrictions
                                    through IDP. The value is the name of the permission
        Returns:
            List of identifiers of App Entity Records that the user can access
        """
        self._verify_role(role)
        self._verify_app_entity_type(app_entity_type)

        user_role = self.users_unit_of_work.user_role_repository.get_first(
            user_id=user.id, role=role
        )
        if not user_role:
            return []

        # Permission restriction get precedence, if existing, for the given app_entity
        if permission:
            permission_restrictions = user_role.permission_restrictions
            if permission_restrictions and permission in permission_restrictions.keys():
                permission_restriction = permission_restrictions.get(permission)
                if permission_app_entity_restriction := permission_restriction.get(
                    app_entity_type
                ):
                    return permission_app_entity_restriction

        # Verify if there is any restriction on the entity for the user
        app_entities_restrictions = user_role.app_entities_restrictions
        if app_entities_restrictions and (
            app_entity_restriction := app_entities_restrictions.get(app_entity_type)
        ):
            return app_entity_restriction

        return ALL



from idp_authentication import SessionPort
from idp_authentication.users.adapters.repositories.user_repository import (
    UserRepository,
)
from idp_authentication.users.adapters.repositories.user_role_repository import (
    UserRoleRepository,
)
from idp_authentication.users.base_classes.base_unit_of_work import BaseUnitOfWork
from idp_authentication.users.domain.ports.unit_of_work import UsersUnitOfWorkPort


class UsersUnitOfWork(BaseUnitOfWork, UsersUnitOfWorkPort):
    def __init__(self, session: SessionPort):
        super().__init__(session)
        self.user_repository = UserRepository(session)
        self.user_role_repository = UserRoleRepository(session)



from enum import Enum, unique
from os import getenv

from pydantic import BaseSettings, Field


def get_postgres_uri(dbname, user, password, host, port=5432):
    return f"postgresql://{user}:{password}@{host}:{port}/{dbname}"


def get_db_url() -> str:
    return get_postgres_uri(
        dbname=getenv("DATABASE_NAME"),
        user=getenv("DATABASE_USER"),
        password=getenv("DATABASE_PASSWORD"),
        host=getenv("DATABASE_HOST"),
    )


def get_test_memory_db_uri():
    return "sqlite:///:memory:"


@unique
class AppEnv(Enum):
    DEV = "development"
    STAGING = "staging"
    PROD = "production"


class Config(BaseSettings):
    APP_IDENTIFIER: str = Field(..., env="APP_IDENTIFIER")
    TENANTS: list = Field(..., env="TENANTS")
    ROLES: list = Field(..., env="ROLES")
    APP_ENTITY_TYPES: list = Field(..., env="APP_ENTITY_TYPES")



import pytest

from idp_authentication import UserRole
from idp_authentication.conftest import (
    TEST_ROLE_1,
    TEST_ROLE_2,
    VEHICLE_APP_ENTITY_IDENTIFIER,
)


@pytest.mark.parametrize(
    "role, app_entities_restrictions, has_access",
    [
        (TEST_ROLE_1, None, True),
        (TEST_ROLE_1, {VEHICLE_APP_ENTITY_IDENTIFIER: None}, True),
        (TEST_ROLE_1, {VEHICLE_APP_ENTITY_IDENTIFIER: [1, 2]}, True),
        (TEST_ROLE_1, {VEHICLE_APP_ENTITY_IDENTIFIER: [3, 4]}, False),
        (TEST_ROLE_1, {"other_entity": [1, 2]}, False),
        (TEST_ROLE_2, {VEHICLE_APP_ENTITY_IDENTIFIER: [1, 2]}, False),
    ],
)
def test_get_users_with_access_to_app_entity_record_use_case(
    container, make_user, role, app_entities_restrictions, has_access
):
    # Arrange
    user = make_user(
        user_roles=[
            UserRole(
                role=role,
                app_entities_restrictions=app_entities_restrictions,
            )
        ]
    )
    record_identifier = 1

    # Act
    user_with_access_to_app = (
        container.users_module()
        .use_cases()
        .get_users_with_access_to_app_entity_record_use_case()
    )
    users = user_with_access_to_app.execute(
        app_entity_type=VEHICLE_APP_ENTITY_IDENTIFIER,
        record_identifier=record_identifier,
        roles=[TEST_ROLE_1],
    )

    # Assert
    assert has_access is (user.id in [u.id for u in users])


def test_get_users_with_access_to_app_entity_record_use_case_does_not_return_inactive_users(
    container, make_user
):
    # Arrange
    record_identifier = 1
    user = make_user(
        is_active=False,
        user_roles=[
            UserRole(
                role=TEST_ROLE_1,
                app_entities_restrictions={
                    VEHICLE_APP_ENTITY_IDENTIFIER: [record_identifier]
                },
            )
        ],
    )

    # Act
    user_with_access_to_app = (
        container.users_module()
        .use_cases()
        .get_users_with_access_to_app_entity_record_use_case()
    )
    users = user_with_access_to_app.execute(
        app_entity_type=VEHICLE_APP_ENTITY_IDENTIFIER,
        record_identifier=record_identifier,
        roles=[TEST_ROLE_1],
    )

    # Assert
    assert user.id not in [u.id for u in users]



def test_create_user(container):
    data = {
        "username": "test0",
        "email": "test0@gmail.com",
        "first_name": "test",
        "last_name": "test",
    }
    create_user_use_case = (
        container.users_module().use_cases().create_or_update_user_use_case()
    )
    created_user = create_user_use_case.execute(data=data)
    assert created_user.username == data["username"]

    users_unit_of_work = container.users_module().users_unit_of_work()
    with users_unit_of_work:
        inserted_user = users_unit_of_work.user_repository.get_or_none(
            username=data["username"]
        )
        assert inserted_user is not None


def test_update_user(container, user):
    data = {
        "username": "test1",
        "email": "test@cardoai.com",
        "first_name": "test",
        "last_name": "test",
    }
    previous_user_email = user.email
    update_user_use_case = (
        container.users_module().use_cases().create_or_update_user_use_case()
    )
    updated_user = update_user_use_case.execute(data=data)
    assert previous_user_email != updated_user.email

    users_unit_of_work = container.users_module().users_unit_of_work()
    with users_unit_of_work as uow:
        db_user = uow.user_repository.get_or_none(username=data["username"])
        assert db_user.email == data["email"]
